#import <Foundation/Foundation.h>

typedef NS_ENUM(NSUInteger, PSHABContactPhoneType) {
    PSHABContactPhoneTypeHome,
    PSHABContactPhoneTypeWork,
    PSHABContactPhoneTypeMain,
    PSHABContactPhoneTypeMobile,
    PSHABContactPhoneTypeIphone,
    PSHABContactPhoneTypeOther
};

/**
 *  Data Access Object for a contact in user's Address Book.
 */
@interface PSHABContactDAO : NSObject

/**
 *  Adress Book record id.
 */
@property (nonatomic, assign) NSInteger abRecordId;

/**
 *  Record order in user's Adress Book.
 */
@property (nonatomic, assign) NSInteger orderId;

@property (nonatomic, strong) NSString* firstName;
@property (nonatomic, strong) NSString* middleName;
@property (nonatomic, strong) NSString* lastName;

@property (nonatomic, assign) NSInteger callingCode;
@property (nonatomic, strong) NSString* phoneNumber;
/**
 *  Phone type specified in user's Adress Book (see `PSHABContactPhoneType`).
 */
@property (nonatomic, assign) PSHABContactPhoneType phoneType;

@property (nonatomic, assign) BOOL isOrganization;
@property (nonatomic, strong) NSString* organization;

/**
 *  Standarized way to show contact's first, middle and last name.
 */
@property (nonatomic, readonly) NSString* fancyName;
/**
 *  Returns the contact phone number with its type -eg: XXX YYY ZZZ (home) -.
 */
@property (nonatomic, readonly) NSString* fancyPhone;
/**
 *  Standarized way to show contact's phone (only numbers).
 */
@property (nonatomic, readonly) NSString* unformattedPhoneNumber;

/**
 *  Returns an unformatted phone number (only numbers) for any formatted phone number.
 *
 *  @param number Input phone number.
 *
 *  @return Unformatted phone number.
 */
+(NSString*)unformattedPhoneNumberForNumber:(NSString*)number;

@end
